open Graph
open Printf
open Ffalgo

type path = string

(* Format of text files:
   % This is a comment

   % A node with its coordinates (which are not used).
   n 88.8 209.7
   n 408.9 183.0

   % The first node has id 0, the next is 1, and so on.

   % Edges: e source dest label
   e 3 1 11
   e 0 2 8

*)

let source = (-1)
let destination = (-2)

let read_factory graph line =
  try Scanf.sscanf line "factory %d %d" (fun id capacity -> new_arc (new_node graph id) source id capacity)
  with e ->
    Printf.printf "Cannot read factory in line - %s:\n%s\n%!" (Printexc.to_string e) line ;
    failwith "from_file"

let read_village graph line =
  try Scanf.sscanf line "village %d %d" (fun id capacity -> new_arc (new_node graph id) id destination capacity)
  with e ->
    Printf.printf "Cannot read village in line - %s:\n%s\n%!" (Printexc.to_string e) line ;
    failwith "from_file"

(* Reads a line with a node. *)
let read_node graph line =
  try Scanf.sscanf line "node %d" (fun id -> new_node graph id)
  with e ->
    Printf.printf "Cannot read node in line - %s:\n%s\n%!" (Printexc.to_string e) line ;
    failwith "from_file"

(* Ensure that the given node exists in the graph. If not, create it. 
 * (Necessary because the website we use to create online graphs does not generate correct files when some nodes have been deleted.) *)
let ensure graph id = if node_exists graph id then graph else new_node graph id

(* Reads a line with an arc. *)
let read_road graph line =
  try Scanf.sscanf line "road %d %d %d"
        (fun id1 id2 capacity -> new_arc (ensure (ensure graph id1) id2) id1 id2 capacity)
  with e ->
    Printf.printf "Cannot read arc in line - %s:\n%s\n%!" (Printexc.to_string e) line ;
    failwith "from_file"

(* Reads a comment or fail. *)
let read_comment graph line =
  try Scanf.sscanf line " %%" graph
  with _ ->
    Printf.printf "Unknown line:\n%s\n%!" line ;
    failwith "from_file"

let circulation_from_file path =

  let infile = open_in path in
  (* create source and destination *)
  let initial_graph = new_node (new_node empty_graph source) destination in
  (* Read all lines until end of file. 
   * n is the current node counter. *)
  let rec loop graph =
    try
      let line = input_line infile in

      (* Remove leading and trailing spaces. *)
      let line = String.trim line in

      let graph2 =
        (* Ignore empty lines *)
        if line = "" then graph

        (* The first character of a line determines its content : n or e. *)
        else match line.[0] with
          | 'f' -> read_factory graph line
          | 'v' -> read_village graph line
          | 'n' -> read_node graph line
          | 'r' -> read_road graph line

          (* It should be a comment, otherwise we complain. *)
          | _ -> read_comment graph line
      in      
      loop graph2

    with End_of_file -> graph (* Done *)
  in

  let final_graph = loop initial_graph in

  close_in infile ;
  {graph = final_graph; origin = source; destination = destination}

